"""
Utilities for doing computation with the oil data

For the most part, what these do is extract the actual data in an Oil
object with specified units, etc.

There are also some facilities for working with the data,
such as interpolating, etc.

"""

pass